/**********************************************************************/
/*   Routines to handle reading/writing to ROM/RAM.		      */
/**********************************************************************/
# include "hp48.h"
# include "uart_io.h"
# include "dis48.h"
# include "proto.h"
# include <string.h>
# include <fcntl.h>
# include <memory.h>

extern long	nibble_masks[];
extern int num_ranges;
extern int	step_flag;
extern int cont_count;
extern CPU cpu;
extern char *blue, *green;

void
read_rom()
{	int	n;
	int	mode = O_RDONLY;
# if defined(O_BINARY)
	mode |= O_BINARY;
# endif
#ifdef MY48
	printf("LOADING BIGROM\n");
 	if ((n = open("bigrom", mode)) < 0) {
 		perror("bigrom");
 		exit(1);
 	}
 
 	if ((cpu.rom = (unsigned char *) malloc(ROM_SIZE)) == (unsigned char *) NULL) {
 		fprintf(stderr, "malloc failed\n");
 		exit(1);
 		}
 	if (read(n, cpu.rom, ROM_SIZE) != ROM_SIZE)
 		printf("Warning: incorrect ROM size\n");
#else
	printf("LOADING ROM\n");
	if ((n = open("rom", mode)) < 0) {
		perror("rom");
		exit(1);
	}

	if ((cpu.rom = (unsigned char *) malloc(ROM_SIZE / 2)) == (unsigned char *) NULL) {
		fprintf(stderr, "malloc failed\n");
		exit(1);
		}
	if (read(n, cpu.rom, ROM_SIZE / 2) != ROM_SIZE / 2)
		printf("Warning: incorrect ROM size\n");
#endif
#ifdef MY48
 	if ((cpu.ram = (Nibble *) malloc(RAM_SIZE*2*sizeof(Nibble)) ) == (Nibble *) NULL) {
 		fprintf(stderr, "Couldn't allocate memory for RAM\n");
 		exit(1);
 		}
 	memset(cpu.ram + 0x100*sizeof(Nibble), 0, 64*sizeof(Nibble));
#else

	if ((cpu.ram = (unsigned char *) malloc(RAM_SIZE)) == (unsigned char *) NULL) {
		fprintf(stderr, "Couldnt allocate memory for RAM\n");
		exit(1);
		}
/*	memset(cpu.ram, 0x54, RAM_SIZE);*/
	memset(cpu.ram + (0x100/2), 0, 64);
#endif
	(void) close(n);
}
/**********************************************************************/
/*   Routine to write a number of nibbles.			      */
/**********************************************************************/
void
write_nibbles(addr, val, num)
long	addr;
long	val;
int	num;
{
	while (num-- > 0) {
		write_nibble(addr++, (int) val);
		val >>= 4;
		}
}
/**********************************************************************/
/*   Routine to write a nibble to RAM.				      */
/**********************************************************************/
void
write_nibble(addr, val)
long	addr;
int	val;
{
	char	*cp;
	int nMustCheckTimer=0;
	addr &= 0xfffffL;
	if (num_ranges && cpu.debug_mode == FALSE)
		check_breakpoint(BP_WRITE, addr);
	/***********************************************/
	/*   Check to see if we're writing to 32K      */
	/*   RAM and we're writing in the correct      */
	/*   place.				       */
	/***********************************************/
	switch ((addr >> 16) & 0x0f) {
	  case 7:
	  	switch (cpu.ram_32k) {
		  case 0x70000L:
		  	break;
		  case 0xFC000L:
		  	if (addr < 0x74000)
				break;
			goto DEFAULT;
		  case 0xFE000L:
		  	if (addr < 0x72000)
				break;
			/* Fall thru... */
		  default:
		  DEFAULT:
		  	printf("%s%05lx: *** BAD WRITE ATTEMPTED AT %05lx%s\n", 
				blue, cpu.pc, addr, green);
		  }
		break;
	  case 0xf:
	  	if (cpu.ram_32k != 0xf0000L) {
			printf("%s%05lx: *** ATTEMPTING TO WRITE TO RAM AT WRONG ADDRESS @%5lx%s\n", 
				blue, cpu.pc, addr, green);
			return;
			}
		/***********************************************/
		/*   Convert hi-ram address to lo-ram	       */
		/*   address.				       */
		/***********************************************/
		addr &= 0x7ffffL;
		break;
	  }
	val &= 0xf;
	/***********************************************/
	/*   Trap writes to certain addresses.	       */
	/***********************************************/
	switch (addr) {
	  case 0x104:
	  	cpu.crc = (cpu.crc & 0xfff0) | val;
		return;
	  case 0x105:
	  	cpu.crc = (cpu.crc & 0xff0f) | (val << 4);
	  	return;
	  case 0x106:
	  	cpu.crc = (cpu.crc & 0xf0ff) | (val << 8);
	  	return;
	  case 0x107:
	  	cpu.crc = (cpu.crc & 0x0fff) | (val << 12);
	  	return;
	  case 0x116:
		if (xmitbufptr<=MAXXMITBUFSIZE)
			xmitbuffer[xmitbufptr++]=val;
		printf("UART OUT %d\n",val);
		return;
	  case 0x137:
	  	cpu.timer1 = val;
		return;
	  case 0x138: case 0x139: case 0x13a: case 0x13b:
	  case 0x13c: case 0x13d: case 0x13e: case 0x13f:
	  	cpu.timer2 &= ~nibble_masks[addr - 0x138];
		cpu.timer2 |= val << ((addr - 0x138)*4);
	  	return;
	  }

#ifdef MY48
 	cpu.ram[addr]=(Nibble)val&0x0f;
#else
  	cp = (char *) &cpu.ram[addr >> 1];
  	if (addr & 1)
  		*cp = (char) ((*cp & 0x0f) | (val << 4));
  	else
  		*cp = (char) ((*cp & 0xf0) | val);
#endif
}

/**********************************************************************/
/*   Routine to read up to 32 bits worth of nibbles.		      */
/**********************************************************************/
long
read_nibbles(addr, len)
long	addr;
int	len;
{	long	val = 0;

	addr += len;
	while (len-- > 0) {
		val = (val << 4) | read_nibble(--addr);
		}
	return val;
}
/**********************************************************************/
/*   Routine to read a byte.					      */
/**********************************************************************/
int
read_byte(addr)
long	addr;
{
	return (read_nibble(addr+1) << 4) | read_nibble(addr);
}
void
calc_crc(val)
int	val;
{
	cpu.crc = (cpu.crc >> 4) ^ (((val ^ cpu.crc) & 0x0f) * 0x1081);
}
/**********************************************************************/
/*   Routine to read a nibble from RAM or ROM.			      */
/**********************************************************************/
int
read_nibble(addr)
long	addr;
{
	int	val;
	int	do_crc = cpu.debug_mode == FALSE;

	addr &= 0xfffffL;

	if (num_ranges && do_crc)
		check_breakpoint(BP_READ, addr);
	/***********************************************/
	/*   Check to see if we're attempting to       */
	/*   read from ROM or RAM in the bank	       */
	/*   switched area.			       */
	/***********************************************/
	switch ((addr >> 16) & 0x0f) {
	  case 0x7:
	  	switch (cpu.ram_32k) {
		  case 0x70000L:
		  	break;
		  case 0xF0000L:
		  	return get_nibble(addr);
		  case 0xFC000L:
		  	if (addr < 0x74000)
				break;
		  	return get_nibble(addr);
		  case 0xFE000L:
		  	if (addr < 0x72000)
				break;
		  	return get_nibble(addr);
		  default:
		  	printf("Help me read this nibble @%05lx (instr=%lx)\n", 
				addr, cpu.instr_count);
		  	return get_nibble(addr);
		  }
		break;
	  case 0xf:
	  	switch (cpu.ram_32k) {
		  case 0xFE000:
		  case 0xFC000:
			break;
		  case 0xF0000:
		  	break;
		  default:
			printf("*** Attempting to read from $%5lx -- do we have a RAM card?\n", addr);
			return 0xf;
		  }
		addr = addr - cpu.ram_32k + 0x70000;
		break;
	  }
	/***********************************************/
	/*   Check for a ROM address.		       */
	/***********************************************/
	if (addr >= 0x00140L && addr < 0x70000L) {
		return get_nibble(addr);
		}

	/***********************************************/
	/*   Check for access to hardware.	       */
	/***********************************************/
	switch (addr) {
	  case 0x104:
	  	return cpu.crc & 0xf;
	  case 0x105:
	  	return (cpu.crc >> 4) & 0xf;
	  case 0x106:
	  	return (cpu.crc >> 8) & 0xf;
	  case 0x107:
	  	return (cpu.crc >> 12) & 0xf;
	  case 0x112:
		return (2|recvavail);
	  case 0x114:
		if (recbufptr<=MAXRECSIZE)
			return recbuffer[recbufptr++];
		else
			return 0;
	  case 0x137:
		return cpu.timer1;
	  case 0x138: case 0x139: case 0x13a: case 0x13b:
	  case 0x13c: case 0x13d: case 0x13e: case 0x13f:
		return (cpu.timer2 >> ((addr - 0x138) * 4)) & 0x0f;
	  }
#ifdef MY48
 	val = cpu.ram[addr];
#else
	if (addr & 1)
		val = (cpu.ram[addr >> 1] >> 4) & 0x0f;
	else
		val = cpu.ram[addr >> 1] & 0x0f;
#endif
	if (do_crc && addr >= 0x200) {
		calc_crc(val);
		}
	return val;
}
/**********************************************************************/
/*   Execute a push of an address.				      */
/**********************************************************************/
void
push(addr)
long	addr;
{
	cpu.rstk_ptr = (cpu.rstk_ptr + 1) & 7;
	cpu.rstk[cpu.rstk_ptr] = addr;
}
/**********************************************************************/
/*   Pop return address.					      */
/**********************************************************************/
long
pop()
{
	long	val = cpu.rstk[cpu.rstk_ptr];
	cpu.rstk[cpu.rstk_ptr] = 0;
	cpu.rstk_ptr = (cpu.rstk_ptr - 1) & 7;
	return val;
}
/**********************************************************************/
/*   Retrieve a variable length immediate constant.		      */
/**********************************************************************/
WORD64 *
get_imm_nibble(addr, len)
long	addr;
int	len;
{	static WORD64	val;
	register int	shift = 0;

	val.hi = 0;
	val.lo = 0;

	while (len-- > 0) {
		if (shift >= 32)
			val.hi |= ((long) get_nibble(addr++)) << (shift-32);
		else
			val.lo |= (long) get_nibble(addr++) << shift;
		shift += 4;
		}
	return &val;
}

