/**********************************************************************/
/*   Routine  to  display  the  LCD  in  an  X-Window. Also supports  */
/*   displaying of the keyboard.				      */
/*   Each  'pixel'  is  represented  as a 2x2 dot to make the screen  */
/*   readable.							      */
/*   								      */
/*   The  code  in  this  file  works with the Athena widget set. It  */
/*   should  be  straightforward to convert to Motif although you'll  */
/*   need to do more work to align the buttons properly.	      */
/*   								      */
/*   Because  the  emulator  is  such  a  CPU  intensive  program we  */
/*   provide  a  number  of  options  for performing keyboard input.  */
/*   Some  are  more  portable  than others. You may need to comment  */
/*   out or change pieces of code for non-BSD derived systems.	      */
/**********************************************************************/
# include	<stdio.h>
# include	<signal.h>
# include	<fcntl.h>
# include	<X11/X.h>
# include	<X11/Intrinsic.h>
# include	<X11/StringDefs.h>
# include	<X11/Shell.h>
# include	<X11/Xaw/Toggle.h>

# include	<X11/Xaw/Box.h>
# include	<X11/Xaw/Paned.h>
# include	<X11/Xaw/Command.h>
# include	"draw.h"
# include "hp48.h"

extern struct stats stats;
extern CPU cpu;

/**********************************************************************/
/*   We  need  to  detect  a  'key'  being pressed and also it being  */
/*   released  with  the mouse. The standard Xaw Command widget only  */
/*   notifies  us  when the button is released. So we need to change  */
/*   the  way  it  works.  First  thing  is  to  define a new action  */
/*   procedure.							      */
/**********************************************************************/
static void	KeyReleased();
static XtActionsRec actions[] = {
  /* {name, procedure}, */
    {"release",		KeyReleased},
    };

/**********************************************************************/
/*   Structure   defining   the  widget  names  and  the  row/column  */
/*   co-ordinates for the keyboard matrix.			      */
/*   								      */
/*   The   table   below  implements  the  following  matrix  layout  */
/*   (copied from Joe Ervins network post).			      */
/*
                IN  #20 #10 #08 #04 #02 #01
 
 
         OUT   (bit)  5   4   3   2   1   0
 
        #100    8         B   C   D   E   F
 
        #080    7       PRG CST VAR  up NXT
                    
        #040    6       STO EVL <<< dwn >>>
 
        #020    5       COS TAN sqt pwr inv
 
        #010    4   ON* ENT +/- EEX DEL <==
 
        #008    3   alp SIN  7   8   9   /
 
        #004    2   yel MTH  4   5   6   x
 
        #002    1   blu  A   1   2   3   -
 
        #001    0        '   0   .  SPC  +
 
 
   (*) The ON key is actually in a column of its own.
   The ON key is represented in bit 15 of the data returned
   from an IN.4 instruction.  #xxx refers to the keyboard scan 
   bits driven with the OUT instruction.   
*/
struct keys {
	char	*name;
	int	code;
	Widget	widget;
	} key_tbl[] = {
	{"A",	0x14},
	{"B",	0x84},
	{"C",	0x83},
	{"D",	0x82},
	{"E",	0x81},
	{"F",	0x80},

	{"MTH",	0x24},
	{"PRG", 0x74},
	{"CST", 0x73},
	{"VAR", 0x72},
	{"UP",	0x71},
	{"NXT",	0x70},

	{"quote",	0x04},
	{"STO",	0x64},
	{"EVAL",	0x63},
	{"LEFT",	0x62},
	{"DOWN",	0x61},
	{"RIGHT",	0x60},

	{"SIN",	0x34},
	{"COS",	0x54},
	{"TAN",	0x53},
	{"SQRT",	0x52},
	{"POW",	0x51},
	{"reciprocal",	0x50},

	{"ENTER",	0x44},
	{"sign",	0x43},
	{"EEX",	0x42},
	{"DEL",	0x41},
	{"backspace",	0x40},

	{"alpha",	0x35},
	{"7",	0x33},
	{"8",	0x32},
	{"9",	0x31},
	{"divide",	0x30},

	{"left",	0x25},
	{"4",	0x23},
	{"5",	0x22},
	{"6",	0x21},
	{"x",	0x20},

	{"right",	0x15},
	{"1",	0x13},
	{"2",	0x12},
	{"3",	0x11},
	{"minus",	0x10},

	{"ATTN",	0x8000},
	{"0",	0x03},
	{"pt",	0x02},
	{"SPC",	0x01},
	{"plus",	0x00},
	{0}
	};
/**********************************************************************/
/*   Flags for the annunciator display.				      */
/**********************************************************************/
# define	ANNUNC_LOC	0x706C3
# define	ANN_BATTERY	0x08
# define	ANN_ALPHA	0x04
# define	ANN_RIGHT	0x02
# define	ANN_LEFT	0x01
# define	ANN_IO		0x20
# define	ANN_BUSY	0x10
struct ann_struct {
	char	*name;
	int	bit;
	Pixmap	pixmap;
	Widget	widget;
	} ann_tbl[] = {
	{"ann_left", ANN_LEFT},
	{"ann_right", ANN_RIGHT},
	{"ann_alpha", ANN_ALPHA},
	{"ann_battery", ANN_BATTERY},
	{"ann_busy", ANN_BUSY},
	{"ann_io", ANN_IO},
	{0}
	};
/**********************************************************************/
/*   Flag set to TRUE to force redraw of annunciators.		      */
/**********************************************************************/
static int	last_annunc_state = -1;

static String fallback_resources[] = {
	"*foreground:		yellow",
	"*background:		black",
	"*BorderColor:		brown",
	"*allowShellResize:	true",
	"*font:		7x13bold",
	"*lcd.foreground:	cyan",
	"*lcd.background:	gray50",

	"*A.width:	38",
	"*B.width:	38",
	"*C.width:	38",
	"*D.width:	38",
	"*E.width:	38",
	"*F.width:	30",

	"*MTH.width:	38",
	"*PRG.width:	38",
	"*CST.width:	38",
	"*VAR.width:	38",
	"*UP.width:	38",
	"*NXT.width:	30",

	"*quote.width:	38",
	"*quote.label:	'",
	"*STO.width:	38",
	"*EVAL.width:	38",
	"*LEFT.width:	38",
	"*DOWN.width:	38",
	"*RIGHT.width:	30",

	"*SIN.width:	38",
	"*COS.width:	38",
	"*TAN.width:	38",
	"*SQRT.width:	38",
	"*POW.width:	38",
	"*reciprocal.width:	30",
	"*reciprocal.label:	1/x",

	"*ENTER.width:	80",
	"*sign.width:	38",
	"*sign.label:	+/-",
	"*EEX.width:	38",
	"*DEL.width:	38",
	"*backspace.width:	30",
	"*backspace.label:	<=",

	"*alpha.width:	43",
	"*7.width:	46",
	"*8.width:	46",
	"*9.width:	46",
	"*divide.width:	48",
	"*divide.label:	/",

	"*left.width:	43",
	"*4.width:	46",
	"*5.width:	46",
	"*6.width:	46",
	"*x.width:	48",

	"*right.width:	43",
	"*1.width:	46",
	"*2.width:	46",
	"*3.width:	46",
	"*minus.width:	48",
	"*minus.label:	-",

	"*ATTN.width:	43",
	"*0.width:	46",
	"*pt.width:	46",
	"*SPC.width:	46",
	"*plus.width:	48",
	"*plus.label:	+",
	(char *) NULL
	};

Widget	top_level;
Widget	ann_box;
Widget	lcd_widget;
XtAppContext	app_con;
GC	gc;
int	x11_loaded = TRUE;
void	kbd_callback();
void	sigio_handler();
void	expose_callback();
void	x11_draw_annunc();
/**********************************************************************/
/*   Remember  state  of  LCD  so  when  we refresh we only draw the  */
/*   bits which are changing.					      */
/**********************************************************************/
# define	ROWS	70
# define	COLUMNS	140
unsigned char	lcd_buffer[ROWS][COLUMNS / 4];

void
x11_init(argcp, argv)
int	*argcp;
char	**argv;
{	int	n;
	Arg	args[20];
	XGCValues	values;
	Display	*dpy;
	Pixel	pixel;
	Widget	pane, kbd;
	int	i;

	n = 0;
	XtSetArg(args[n], XtNinput, TRUE); n++;
	XtSetArg(args[n], XtNallowShellResize, TRUE); n++;
	top_level = XtAppInitialize(&app_con, "HP48", NULL, 0,
		argcp, argv, fallback_resources, args, n);
	dpy = XtDisplay(top_level);

	/***********************************************/
	/*   Register our private actions table.       */
	/***********************************************/
	XtAppAddActions(app_con, actions, XtNumber(actions));

	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	pane = XtCreateWidget("pane", panedWidgetClass,
			top_level, args, n);
	XtManageChild(pane);

	/***********************************************/
	/*   Create  box  at  top  of display for the  */
	/*   annunciator labels.		       */
	/***********************************************/
	n = 0;
	XtSetArg(args[n], XtNheight, 8); n++;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	ann_box = XtCreateManagedWidget("annunc", boxWidgetClass, pane, args, n);

	for (i = 0; ann_tbl[i].name; i++) {
		n = 0;
		ann_tbl[i].widget = XtCreateManagedWidget(ann_tbl[i].name, 
			labelWidgetClass, ann_box, args, n);
		/***********************************************/
		/*   Get  the  pixmap created for us from the  */
		/*   defaults file.			       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNbitmap, &ann_tbl[i].pixmap); n++;
		XtGetValues(ann_tbl[i].widget, args, n);
		/***********************************************/
		/*   Now blank out the annunciator.	       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNlabel, ""); n++;
		XtSetValues(ann_tbl[i].widget, args, n);
		}

	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	XtSetArg(args[n], XtNheight, 64 * 2); n++;
	XtSetArg(args[n], XtNwidth, 132 * 2); n++;
	lcd_widget = XtCreateManagedWidget("lcd", drawWidgetClass, pane, args, n);
	XtAddCallback(lcd_widget, XtNexposeCallback, expose_callback, NULL);

	n = 0;
	XtSetArg(args[n], XtNheight, 64 * 2); n++;
	XtSetArg(args[n], XtNwidth, 132 * 2); n++;
	kbd = XtCreateManagedWidget("keyboard", boxWidgetClass, pane, args, n);

	XawPanedAllowResize(kbd, TRUE);

	for (i = 0; key_tbl[i].name; i++) {
		n = 0;
		key_tbl[i].widget = XtCreateManagedWidget(key_tbl[i].name, commandWidgetClass, kbd, args, n);
		XtAddCallback(key_tbl[i].widget, XtNcallback, kbd_callback, key_tbl[i].code);
		}

	XtRealizeWidget(top_level);

	gc = XCreateGC(dpy, XtWindow(lcd_widget), 0, &values);
	n = 0;
	XtSetArg(args[n], XtNforeground, &pixel); n++;
	XtGetValues(lcd_widget, args, n);
	XSetForeground(XtDisplay(top_level), gc, pixel);

	XFlush(XtDisplay(top_level));

}
/**********************************************************************/
/*   Routine to draw a nibble into the display.			      */
/**********************************************************************/
void
x11_draw_nibble(x, y, val)
int	x;
int	y;
int	val;
{	XPoint	pt[4 * 4];
	XPoint	*ptp = pt;
	int	nibble;
	int	i, r, c, b;

	if (y > ROWS || x > COLUMNS / 4)
		return;
	c = x * 4;
	b = 1;
	r = y;

	x *= 4 * 2;
	y *= 2;

	nibble = lcd_buffer[r][c];
	for (i = 0; i < 4; i++, val >>= 1) {
		if (val & 1) {
			if ((nibble & b) == 0) {
				nibble |= b;
				ptp[0].x = x; 	ptp[0].y = y;
				ptp[1].x = x+1; ptp[1].y = y;
				ptp[2].x = x; 	ptp[2].y = y+1;
				ptp[3].x = x+1; ptp[3].y = y+1;
				ptp += 4;
				}
			}
		else {
			if (nibble & b) {
				nibble &= ~b;
				XClearArea(XtDisplay(lcd_widget), XtWindow(lcd_widget),
					x, y, 2, 2, False);
				}
			}
		b <<= 1;
		x += 2;
		}
	lcd_buffer[r][c] = nibble;
	if (ptp != pt)
		XDrawPoints(XtDisplay(lcd_widget), XtWindow(lcd_widget), 
			gc, pt, ptp - pt, CoordModeOrigin);
}
void
x11_flush()
{
	XFlush(XtDisplay(top_level));
}
void
x11_clear()
{	int	flags;

	memset(lcd_buffer, 0, sizeof lcd_buffer);
	XClearWindow(XtDisplay(lcd_widget), XtWindow(lcd_widget));
	/***********************************************/
	/*   Fill in status of annunicators.	       */
	/***********************************************/
	flags = read_nibbles(ANNUNC_LOC, 2);
	if (flags & ANN_ALPHA) {
		}
	x11_flush();
}
/**********************************************************************/
/*   Function  called  from  debugger command line to refresh window  */
/*   and handle any pending events.				      */
/**********************************************************************/
void
x11_refresh_window()
{	XEvent	event;
	
	while (XtAppPending(app_con)) {
		XtAppNextEvent(app_con, &event);
		XtDispatchEvent(&event);
		}
	x11_draw_annunc();
}
/**********************************************************************/
/*   Routine  called  to  turn  on/off async (SIGIO) notification of  */
/*   activity   on  X  file-descriptor  so  we  can  handle  repaint  */
/*   messages  etc  whilst  in  heavy emulation mode without slowing  */
/*   down the core processing module.				      */
/**********************************************************************/
void
x11_set_async(flag)
int	flag;
{
	/***********************************************/
	/*   Following  code  is  OPTIONAL.  We allow  */
	/*   SIGIO  signals  to  go off on the X file  */
	/*   descriptor  to  attract  our  attention.  */
	/*   We  just  set a global flag. This causes  */
	/*   the  code  in  execute_loop() to look to  */
	/*   see   what  happened  and  if  necessary  */
	/*   look  at  the  X input queue and process  */
	/*   any  events.  This  is  obviously not as  */
	/*   portable   as   the  other  methods  but  */
	/*   hopefully    gives   us   a   guaranteed  */
	/*   response.   If   this   signal   is  not  */
	/*   enabled  then  you'll  have  to  rely on  */
	/*   the   32768   instruction   sampler  for  */
	/*   looking  for  input  or the execution of  */
	/*   a SHUTDN instruction.		       */
	/***********************************************/
# if defined(SIGIO) && defined(FASYNC) && defined(F_SETOWN)
	{int flags;
	int	fd;

	fd = XtDisplay(top_level)->fd;
	siginterrupt(SIGIO, FALSE);
	signal(SIGIO, sigio_handler);
	flags = fcntl(fd, F_GETFL, 0);
	if (flag)
		fcntl(fd, F_SETFL, flags | FASYNC);
	else
		fcntl(fd, F_SETFL, flags & ~FASYNC);
	fcntl(fd, F_SETOWN, getpid());
	}
# endif
}
/**********************************************************************/
/*   Routine to draw the annunciators.				      */
/**********************************************************************/
void
x11_draw_annunc()
{	int	n1, n2;
	int	n, i;
	int	f;
	Arg	args[20];

	n1 = read_nibble(0x10B);
	n2 = read_nibble(0x10C);
	f = (n2 << 4) | n1;
	if (f == last_annunc_state)
		return;
	last_annunc_state = f;
	for (i = 0; ann_tbl[i].name; i++) {
		n = 0;
		if (ann_tbl[i].bit & f) {
			XtSetArg(args[n], XtNbitmap, ann_tbl[i].pixmap); n++;
			}
		else {
			XtSetArg(args[n], XtNbitmap, None); n++;
			XtSetArg(args[n], XtNlabel, ""); n++;
			}
		XtSetValues(ann_tbl[i].widget, args, n);
		}
}
/**********************************************************************/
/*   Callback  routine  when  user  clicks  on  a keyboard key. Just  */
/*   provoke  the  emulator  into  believing  the  key specified was  */
/*   pressed.							      */
/**********************************************************************/
void
kbd_callback(w, code, ptr)
Widget	w;
int	code;
caddr_t	ptr;
{	int	r, c;
	int	i;
	extern FILE	*cmd_fp;

	/***********************************************/
	/*   Handle the ATTN key special.	       */
	/***********************************************/
	if (code == 0x8000) {
		do_intr(0, (long *) NULL);
		for (i = 0; i < 9; i++)
			cpu.keybuf.rows[i] ^= 0x8000;
		return;
		}
	r = code >> 4;
	c = 1 << (code & 0x0f);
	/***********************************************/
	/*   Generate an interrupt if key going down.  */
	/***********************************************/
	if ((cpu.keybuf.rows[r] & c) == 0) {
		if (cpu.ien)
			do_intr(0, (long *) NULL);
		cpu.keybuf.rows[r] |= c;
		/***********************************************/
		/*   Log command at exact time it occured.     */
		/***********************************************/
		fprintf(cmd_fp, "instr %lx\nc\nkey %x\n", 
			cpu.instr_count, code);
		}
	else {
		cpu.keybuf.rows[r] &= ~c;
		/***********************************************/
		/*   Turn off ATTN key as well.		       */
		/***********************************************/
		for (i = 0; i < 9; i++)
			cpu.keybuf.rows[i] &= ~0x8000;
		}
	
}
/**********************************************************************/
/*   Function  called  when  SIGIO signal goes off indicating some X  */
/*   input    is   available.   Just   set   a   flag   sampled   by  */
/*   execute_loop().						      */
/**********************************************************************/
void
sigio_handler()
{
	exec_flags |= EXEC_SIGIO;
}
/**********************************************************************/
/*   Routine  called  when  we receive an expose event. Just repaint  */
/*   the LCD.							      */
/**********************************************************************/
void
expose_callback()
{	int	n;
	XEvent ev;

	/***********************************************/
	/*   Try and merge all exposure events.	       */
	/***********************************************/
	while (XCheckWindowEvent(XtDisplay(lcd_widget), XtWindow(lcd_widget), ExposureMask, &ev))
		;
	/***********************************************/
	/*   Expose   event   if   LCD   hasnt   been  */
	/*   initialised.			       */
	/***********************************************/
	n = read_nibble(0x124);
	if (n != 0x7 && n != 0xF)
		return;
	memset(lcd_buffer, 0, sizeof lcd_buffer);
	do_lcd();
}
/**********************************************************************/
/*   Action procedure called when a key is released.		      */
/**********************************************************************/
static void
KeyReleased(w, event, x, y)
Widget		w;
XEvent		*event;
String		*x;
Cardinal	*y;
{
	memset(&cpu.keybuf, 0, sizeof cpu.keybuf);
}
